import { defineStore } from "pinia";
import type { User } from "@/client/auth";
import { UserService, RoleEnum } from "@/client/auth";
import { OpenAPI as S3ProxyOpenAPI } from "@/client/s3proxy";
import { OpenAPI as AuthOpenAPI } from "@/client/auth";
import { OpenAPI as WorkflowOpenAPI } from "@/client/workflow";
import { OpenAPI as ResourceOpenAPI } from "@/client/resource";
import { useWorkflowExecutionStore } from "@/stores/workflowExecutions";
import { useBucketStore } from "@/stores/buckets";
import { useWorkflowStore } from "@/stores/workflows";
import { useS3KeyStore } from "@/stores/s3keys";
import { useS3ObjectStore } from "@/stores/s3objects";
import { clear as dbclear } from "idb-keyval";
import { useNameStore } from "@/stores/names";

type DecodedToken = {
  exp: number;
  iss: string;
  roles: RoleEnum[];
  sub: string;
};

export type RootState = {
  token: string | null;
  decodedToken: DecodedToken | null;
  user: User | null;
};

function parseJwt(token: string): DecodedToken {
  const base64Url = token.split(".")[1];
  const base64 = base64Url.replace(/-/g, "+").replace(/_/g, "/");
  const jsonPayload = decodeURIComponent(
    window
      .atob(base64)
      .split("")
      .map(function (c) {
        return "%" + ("00" + c.charCodeAt(0).toString(16)).slice(-2);
      })
      .join(""),
  );

  return JSON.parse(jsonPayload) as DecodedToken;
}

export const useAuthStore = defineStore({
  id: "auth",
  state: () =>
    ({
      token: null,
      decodedToken: null,
      user: null,
    }) as RootState,
  getters: {
    roles(): string[] {
      return (
        this.user?.roles?.map((role) => role.toString()) ??
        this.decodedToken?.["roles"] ??
        []
      );
    },
    authenticated: (state) => state.token != null,
    currentUID(): string {
      return this.decodedToken?.["sub"] ?? "";
    },
    foreignUser: (state) =>
      state.user?.roles?.includes(RoleEnum.FOREIGN_USER) ??
      state.decodedToken?.roles.includes(RoleEnum.FOREIGN_USER) ??
      true,
    normalUser: (state) =>
      state.user?.roles?.includes(RoleEnum.USER) ??
      state.decodedToken?.roles.includes(RoleEnum.USER) ??
      false,
    rewiewer: (state) =>
      state.user?.roles?.includes(RoleEnum.REVIEWER) ??
      state.decodedToken?.roles.includes(RoleEnum.REVIEWER) ??
      false,
    workflowDev: (state) =>
      state.user?.roles?.includes(RoleEnum.DEVELOPER) ??
      state.decodedToken?.roles.includes(RoleEnum.DEVELOPER) ??
      false,
    resourceMaintainer: (state) =>
      state.user?.roles?.includes(RoleEnum.DB_MAINTAINER) ??
      state.decodedToken?.roles.includes(RoleEnum.DB_MAINTAINER) ??
      false,
    admin: (state) =>
      state.user?.roles?.includes(RoleEnum.ADMINISTRATOR) ??
      state.decodedToken?.roles.includes(RoleEnum.ADMINISTRATOR) ??
      false,
  },
  actions: {
    setToken(token: string | null) {
      if (token != null) {
        this.token = token;
        this.decodedToken = parseJwt(token);
        S3ProxyOpenAPI.TOKEN = token;
        AuthOpenAPI.TOKEN = token;
        WorkflowOpenAPI.TOKEN = token;
        ResourceOpenAPI.TOKEN = token;
        UserService.userGetLoggedInUser()
          .then((user) => {
            this.updateUser(user);
          })
          .catch(() => {
            this.token = null;
          });
      } else {
        this.logout();
      }
    },
    updateUser(user: User) {
      this.user = user;
      useNameStore().addNameToMapping(user.uid, user.display_name);
    },
    logout() {
      S3ProxyOpenAPI.TOKEN = undefined;
      AuthOpenAPI.TOKEN = undefined;
      WorkflowOpenAPI.TOKEN = undefined;
      ResourceOpenAPI.TOKEN = undefined;
      this.$reset();
      localStorage.clear();
      dbclear();
      useWorkflowExecutionStore().$reset();
      useBucketStore().$reset();
      useWorkflowStore().$reset();
      useS3KeyStore().$reset();
      useS3ObjectStore().$reset();
    },
    async fetchUsernames(uids: string[]): Promise<string[]> {
      const nameStore = useNameStore();
      const filteredIds = uids
        .filter((uid) => !nameStore.getName(uid)) // filter already present UIDs
        .filter(
          // filter unique UIDs
          (modeId, index, array) =>
            array.findIndex((val) => val === modeId) === index,
        );
      // If all uids are already in the store, then return them
      if (filteredIds.length === 0) {
        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
        return uids.map((uid) => nameStore.getName(uid)!);
      }
      const missingIds: string[] = [];
      const storedNames = filteredIds.map((uid) => localStorage.getItem(uid));

      for (const index in storedNames) {
        // if uid was not cached, mark it to fetch it from backend
        if (storedNames[index] == null) {
          missingIds.push(filteredIds[index]);
        } else {
          // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
          nameStore.addNameToMapping(filteredIds[index], storedNames[index]!);
        }
      }
      // fetch missing users from backend
      const fetchedUsers = await Promise.all(
        missingIds.map((uid) => UserService.userGetUser(uid)),
      );
      // Put users in store
      for (const user of fetchedUsers) {
        nameStore.addNameToMapping(user.uid, user.display_name);
      }
      // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
      return uids.map((uid) => nameStore.getName(uid)!);
    },
  },
});
