import { defineStore } from "pinia";
import type {
  Body_Workflow_Version_upload_workflow_version_icon,
  IconUpdateOut,
  WorkflowCredentialsIn,
  WorkflowIn,
  WorkflowModeOut,
  WorkflowOut,
  WorkflowUpdate,
  WorkflowVersion,
} from "@/client/workflow";
import {
  Status,
  WorkflowCredentialsService,
  WorkflowModeService,
  WorkflowService,
  WorkflowVersionService,
} from "@/client/workflow";
import { useAuthStore } from "@/stores/users";
import { set, get } from "idb-keyval";
import { useNameStore } from "@/stores/names";

export const useWorkflowStore = defineStore({
  id: "workflows",
  state: () =>
    ({
      workflowMapping: {},
      comprehensiveWorkflowMapping: {},
      modeMapping: {},
    }) as {
      workflowMapping: Record<string, WorkflowOut>;
      comprehensiveWorkflowMapping: Record<string, WorkflowOut>;
      modeMapping: Record<string, WorkflowModeOut>;
    },
  getters: {
    workflows(): WorkflowOut[] {
      return Object.values(this.workflowMapping);
    },
    ownWorkflows(): WorkflowOut[] {
      const authStore = useAuthStore();
      return Object.values(this.comprehensiveWorkflowMapping).filter(
        (workflow) => workflow.developer_id === authStore.currentUID,
      );
    },
    reviewableWorkflows(): WorkflowOut[] {
      return Object.values(this.comprehensiveWorkflowMapping).filter(
        (workflow) =>
          workflow.versions.filter(
            (version) => version.status === Status.CREATED,
          ).length > 0,
      );
    },
    versionMapping(): Record<string, WorkflowVersion> {
      const mapping: Record<string, WorkflowVersion> = {};
      for (const workflow of this.workflows) {
        for (const version of workflow.versions) {
          mapping[version.git_commit_hash] = version;
        }
      }
      return mapping;
    },
    getArbitraryWorkflow(): (wid: string) => Promise<WorkflowIn | undefined> {
      return (wid: string) => get(wid);
    },
  },
  actions: {
    __addNameToMapping(key: string, value: string) {
      useNameStore().addNameToMapping(key, value);
    },
    fetchWorkflows(onFinally?: () => void): Promise<WorkflowOut[]> {
      if (Object.keys(this.workflowMapping).length > 0) {
        onFinally?.();
      }
      return WorkflowService.workflowListWorkflows()
        .then((workflows) => {
          for (const workflow of workflows) {
            this.workflowMapping[workflow.workflow_id] = workflow;
            this.__addNameToMapping(workflow.workflow_id, workflow.name);
            for (const version of workflow.versions) {
              this.__addNameToMapping(version.git_commit_hash, version.version);
            }
          }
          this.fetchWorkflowModes(
            workflows
              .map((workflow) => workflow.versions ?? [])
              .flat()
              .map((version) => version.modes ?? [])
              .flat(),
          );
          return workflows;
        })
        .finally(onFinally);
    },
    fetchOwnWorkflows(onFinally?: () => void): Promise<WorkflowOut[]> {
      const authStore = useAuthStore();
      if (this.ownWorkflows.length > 0) {
        onFinally?.();
      }
      return WorkflowService.workflowListWorkflows(
        undefined,
        Object.values(Status),
        authStore.currentUID,
      )
        .then((workflows) => {
          for (const workflow of workflows) {
            this.comprehensiveWorkflowMapping[workflow.workflow_id] = workflow;
            this.__addNameToMapping(workflow.workflow_id, workflow.name);
            for (const version of workflow.versions) {
              this.__addNameToMapping(version.git_commit_hash, version.version);
            }
          }
          this.fetchWorkflowModes(
            workflows
              .map((workflow) => workflow.versions ?? [])
              .flat()
              .map((version) => version.modes ?? [])
              .flat(),
          );
          return workflows;
        })
        .finally(onFinally);
    },
    fetchReviewableWorkflows(onFinally?: () => void): Promise<WorkflowOut[]> {
      if (this.reviewableWorkflows.length > 0) {
        onFinally?.();
      }
      return WorkflowService.workflowListWorkflows(undefined, [Status.CREATED])
        .then((workflows) => {
          for (const workflow of workflows) {
            this.__addNameToMapping(workflow.workflow_id, workflow.name);
            for (const version of workflow.versions) {
              this.__addNameToMapping(version.git_commit_hash, version.version);
            }
            if (
              this.comprehensiveWorkflowMapping[workflow.workflow_id] !=
              undefined
            ) {
              // merge cached workflow versions and fetched workflow versions
              for (const version of workflow.versions) {
                // if version is not present, push version into versions list
                if (
                  this.comprehensiveWorkflowMapping[
                    workflow.workflow_id
                  ].versions.findIndex(
                    (workflowVersion) =>
                      workflowVersion.git_commit_hash ===
                      version.git_commit_hash,
                  ) < 0
                ) {
                  this.comprehensiveWorkflowMapping[
                    workflow.workflow_id
                  ].versions.push(version);
                }
              }
            } else {
              this.comprehensiveWorkflowMapping[workflow.workflow_id] =
                workflow;
            }
          }
          return workflows;
        })
        .finally(onFinally);
    },
    fetchWorkflow(
      workflow_id: string,
      comprehensive = false,
      onFinally?: () => void,
    ): Promise<WorkflowOut> {
      if (
        comprehensive &&
        this.comprehensiveWorkflowMapping[workflow_id] != undefined
      ) {
        onFinally?.();
      }
      if (!comprehensive && this.workflowMapping[workflow_id] != undefined) {
        onFinally?.();
      }
      return WorkflowService.workflowGetWorkflow(
        workflow_id,
        comprehensive ? Object.values(Status) : undefined,
      )
        .then((workflow) => {
          this.__addNameToMapping(workflow.workflow_id, workflow.name);
          for (const version of workflow.versions) {
            this.__addNameToMapping(version.git_commit_hash, version.version);
          }
          if (comprehensive) {
            this.comprehensiveWorkflowMapping[workflow_id] = workflow;
          } else {
            this.workflowMapping[workflow_id] = workflow;
          }
          this.fetchWorkflowModes(
            workflow.versions.map((version) => version.modes ?? []).flat(),
          );
          return workflow;
        })
        .finally(onFinally);
    },
    fetchWorkflowModes(modeIds?: string[]): Promise<WorkflowModeOut[]> {
      if (modeIds == undefined) {
        return Promise.resolve([]);
      }
      const filteredIds = modeIds
        .filter((modeId) => !this.modeMapping[modeId]) // filter null modes and already present modes
        .filter(
          // filter unique workflow versions
          (modeId, index, array) =>
            array.findIndex((val) => val === modeId) === index,
        );
      return Promise.all(
        filteredIds.map((modeId) =>
          WorkflowModeService.workflowModeGetWorkflowMode(modeId),
        ),
      ).then((modes) => {
        for (const mode of modes) {
          this.__addNameToMapping(mode.mode_id, mode.name);
          this.modeMapping[mode.mode_id] = mode;
        }
        return modes;
      });
    },
    async setArbitraryWorkflow(workflow: WorkflowIn) {
      const wid = crypto.randomUUID();
      await set(wid, workflow);
      return wid;
    },
    deprecateWorkflowVersion(
      workflow_id: string,
      version_id: string,
    ): Promise<WorkflowVersion> {
      return WorkflowVersionService.workflowVersionDeprecateWorkflowVersion(
        workflow_id,
        version_id,
      ).then((version) => {
        // Update version in workflowMapping
        if (this.workflowMapping[workflow_id] == undefined) {
          this.fetchWorkflow(workflow_id, false);
        } else {
          const versionIndex1 = this.workflowMapping[
            workflow_id
          ].versions.findIndex(
            (version) => version.git_commit_hash == version_id,
          );
          if (versionIndex1 > -1) {
            this.workflowMapping[workflow_id].versions[versionIndex1] = version;
          } else {
            this.workflowMapping[workflow_id].versions.push(version);
          }
        }

        // Update version in comprehensiveWorkflowMapping
        if (this.comprehensiveWorkflowMapping[workflow_id] == undefined) {
          this.fetchWorkflow(workflow_id, true);
        } else {
          const versionIndex2 = this.comprehensiveWorkflowMapping[
            workflow_id
          ].versions.findIndex(
            (version) => version.git_commit_hash == version_id,
          );
          if (versionIndex2 > -1) {
            this.comprehensiveWorkflowMapping[workflow_id].versions[
              versionIndex2
            ] = version;
          } else {
            this.comprehensiveWorkflowMapping[workflow_id].versions.push(
              version,
            );
          }
        }
        return version;
      });
    },
    updateWorkflow(
      workflow_id: string,
      version: WorkflowUpdate,
    ): Promise<WorkflowVersion> {
      return WorkflowService.workflowUpdateWorkflow(workflow_id, version).then(
        (updatedVersion) => {
          this.__addNameToMapping(
            updatedVersion.git_commit_hash,
            updatedVersion.version,
          );
          if (this.comprehensiveWorkflowMapping[workflow_id] == undefined) {
            this.fetchWorkflow(workflow_id, true);
          } else {
            this.comprehensiveWorkflowMapping[workflow_id].versions.push(
              updatedVersion,
            );
          }
          return updatedVersion;
        },
      );
    },
    deleteWorkflow(workflow_id: string): Promise<void> {
      return WorkflowService.workflowDeleteWorkflow(workflow_id).then(() => {
        useNameStore().deleteNameFromMapping(workflow_id);
        delete this.workflowMapping[workflow_id];
        delete this.comprehensiveWorkflowMapping[workflow_id];
      });
    },
    createWorkflow(workflow: WorkflowIn): Promise<WorkflowOut> {
      return WorkflowService.workflowCreateWorkflow(workflow).then(
        (workflowOut) => {
          this.comprehensiveWorkflowMapping[workflowOut.workflow_id] =
            workflowOut;
          this.__addNameToMapping(workflowOut.workflow_id, workflowOut.name);
          this.__addNameToMapping(
            workflow.git_commit_hash,
            workflowOut.versions[0].version,
          );
          return workflowOut;
        },
      );
    },
    updateWorkflowCredentials(
      workflow_id: string,
      credentials: WorkflowCredentialsIn,
    ): Promise<void> {
      return WorkflowCredentialsService.workflowCredentialsUpdateWorkflowCredentials(
        workflow_id,
        credentials,
      ).then(() => {
        if (this.workflowMapping[workflow_id] == undefined) {
          this.fetchWorkflow(workflow_id, false);
        } else {
          this.workflowMapping[workflow_id].private = true;
        }
        if (this.comprehensiveWorkflowMapping[workflow_id] == undefined) {
          this.fetchWorkflow(workflow_id, true);
        } else {
          this.comprehensiveWorkflowMapping[workflow_id].private = true;
        }
      });
    },
    deleteWorkflowCredentials(workflow_id: string): Promise<void> {
      return WorkflowCredentialsService.workflowCredentialsDeleteWorkflowCredentials(
        workflow_id,
      ).then(() => {
        if (this.workflowMapping[workflow_id] == undefined) {
          this.fetchWorkflow(workflow_id, false);
        } else {
          this.workflowMapping[workflow_id].private = false;
        }
        if (this.comprehensiveWorkflowMapping[workflow_id] == undefined) {
          this.fetchWorkflow(workflow_id, true);
        } else {
          this.comprehensiveWorkflowMapping[workflow_id].private = false;
        }
      });
    },
    updateVersionIcon(
      workflow_id: string,
      version_id: string,
      icon: Body_Workflow_Version_upload_workflow_version_icon,
    ): Promise<IconUpdateOut> {
      return WorkflowVersionService.workflowVersionUploadWorkflowVersionIcon(
        workflow_id,
        version_id,
        icon,
      ).then((response) => {
        // Update version in workflowMapping
        if (this.workflowMapping[workflow_id] == undefined) {
          this.fetchWorkflow(workflow_id, false);
        } else {
          const versionIndex1 = this.workflowMapping[
            workflow_id
          ].versions.findIndex(
            (version) => version.git_commit_hash == version_id,
          );
          if (versionIndex1 > -1) {
            this.workflowMapping[workflow_id].versions[versionIndex1].icon_url =
              response.icon_url;
          }
        }

        // Update version in comprehensiveWorkflowMapping
        if (this.comprehensiveWorkflowMapping[workflow_id] == undefined) {
          this.fetchWorkflow(workflow_id, true);
        } else {
          const versionIndex2 = this.comprehensiveWorkflowMapping[
            workflow_id
          ].versions.findIndex(
            (version) => version.git_commit_hash == version_id,
          );
          if (versionIndex2 > -1) {
            setTimeout(() => {
              this.comprehensiveWorkflowMapping[workflow_id].versions[
                versionIndex2
              ].icon_url = response.icon_url;
            }, 500);
          }
        }
        return response;
      });
    },
    deleteVersionIcon(workflow_id: string, version_id: string): Promise<void> {
      return WorkflowVersionService.workflowVersionDeleteWorkflowVersionIcon(
        workflow_id,
        version_id,
      ).then(() => {
        // Update version in workflowMapping
        if (this.workflowMapping[workflow_id] == undefined) {
          this.fetchWorkflow(workflow_id, false);
        } else {
          const versionIndex1 = this.workflowMapping[
            workflow_id
          ].versions.findIndex(
            (version) => version.git_commit_hash == version_id,
          );
          if (versionIndex1 > -1) {
            this.workflowMapping[workflow_id].versions[versionIndex1].icon_url =
              null;
          }
        }

        // Update version in comprehensiveWorkflowMapping
        if (this.comprehensiveWorkflowMapping[workflow_id] == undefined) {
          this.fetchWorkflow(workflow_id, true);
        } else {
          const versionIndex2 = this.comprehensiveWorkflowMapping[
            workflow_id
          ].versions.findIndex(
            (version) => version.git_commit_hash == version_id,
          );
          if (versionIndex2 > -1) {
            this.comprehensiveWorkflowMapping[workflow_id].versions[
              versionIndex2
            ].icon_url = null;
          }
        }
      });
    },
    updateVersionStatus(
      workflow_id: string,
      version_id: string,
      status: Status,
    ): Promise<void> {
      return WorkflowVersionService.workflowVersionUpdateWorkflowVersionStatus(
        version_id,
        workflow_id,
        { status: status },
      ).then(() => {
        if (this.comprehensiveWorkflowMapping[workflow_id] == undefined) {
          this.fetchWorkflow(workflow_id, true);
        } else {
          const versionIndex = this.comprehensiveWorkflowMapping[
            workflow_id
          ].versions.findIndex(
            (version) => version.git_commit_hash == version_id,
          );
          if (versionIndex > -1) {
            this.comprehensiveWorkflowMapping[workflow_id].versions[
              versionIndex
            ].status = status;
          }
        }
      });
    },
  },
});
